package input;

import java.io.IOException;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.InputSplit;
import org.apache.hadoop.mapreduce.RecordReader;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.mapreduce.lib.input.FileSplit;
import org.apache.hadoop.util.LineReader;

public class GeneFamilyRecordReader extends RecordReader<LongWritable, GeneFamily> {

	private LongWritable key;
	private GeneFamily value;
	private LineReader in=null;
	
	
	private long start =0;
    private long end =0;
    private long pos =0;
	private int maxLineLength;
	
	@Override
	public void close() throws IOException {
		if (in != null) {
            in.close();
        }
		
	}

	@Override
	public LongWritable getCurrentKey() throws IOException,
			InterruptedException {
		return key;
	}

	@Override
	public GeneFamily getCurrentValue() throws IOException,
			InterruptedException {
		return value;
	}

	@Override
	public float getProgress() throws IOException, InterruptedException {
		if (start == end) {
            return 0.0f;
        }
        else {
            return Math.min(1.0f, (pos - start) / (float)(end - start));
        }
	}
	

	
	/*
	 * initialize split variables
	 */
	@Override
	public void initialize(InputSplit genericSplit, TaskAttemptContext context)
			throws IOException, InterruptedException {
		
        // This InputSplit is a FileInputSplit
        FileSplit split = (FileSplit) genericSplit;
		
        // Retrieve configuration, and Max allowed
        // bytes for a single record
        Configuration job = context.getConfiguration();
        this.maxLineLength = job.getInt(
                "mapred.linerecordreader.maxlength",
                Integer.MAX_VALUE);
 
        // Split "S" is responsible for all records
        // starting from "start" and "end" positions
        start = split.getStart();
        end = start + split.getLength();
 
        // Retrieve file containing Split "S"
        final Path file = split.getPath();
        FileSystem fs = file.getFileSystem(job);
        FSDataInputStream fileIn = fs.open(split.getPath());
 
        // If Split "S" starts at byte 0, first line will be processed
        // If Split "S" does not start at byte 0, first line has been already
        // processed by "S-1" and therefore needs to be silently ignored
        boolean skipFirstLine = false;
        if (start != 0) {
            skipFirstLine = true;
            // Set the file pointer at "start - 1" position.
            // This is to make sure we won't miss any line
            // It could happen if "start" is located on a EOL
            --start;
            fileIn.seek(start);
        }
 
        in = new LineReader(fileIn, job);
 
        // If first line needs to be skipped, read first line
        // and stores its content to a dummy Text
        if (skipFirstLine) {
            Text dummy = new Text();
            // Reset "start" to "start + line offset"
            start += in.readLine(dummy, 0,     		
            		maxBytesToConsume(end-start));
                    
        }
 
        // Position is the actual start
        this.pos = start;
 	
	}
	
	private int maxBytesToConsume(long endMinusStart){
		return (int)Math.min((long)Integer.MAX_VALUE, endMinusStart);
	}

	

	/* COPY from blog for custon record reader
	 @Override
    public boolean nextKeyValue() throws IOException {
 
        // Current offset is the key
        key.set(pos);
 
        int newSize = 0;
 
        // Make sure we get at least one record that starts in this Split
        while (pos < end) {
 
            // Read first line and store its content to "value"
            newSize = in.readLine(value, maxLineLength,
                    Math.max((int) Math.min(
                            Integer.MAX_VALUE, end - pos),
                            maxLineLength));
 
            // No byte read, seems that we reached end of Split
            // Break and return false (no key / value)
            if (newSize == 0) {
                break;
            }
 
            // Line is read, new position is set
            pos += newSize;
 
            // Line is lower than Maximum record line size
            // break and return true (found key / value)
            if (newSize < maxLineLength) {
                break;
            }
 
            // Line is too long
            // Try again with position = position + line offset,
            // i.e. ignore line and go to next one
            // TODO: Shouldn't it be LOG.error instead ??
            LOG.info("Skipped line of size " +
                    newSize + " at pos "
                    + (pos - newSize));
        }
 
         
        if (newSize == 0) {
            // We've reached end of Split
            key = null;
            value = null;
            return false;
        } else {
            // Tell Hadoop a new line has been found
            // key / value will be retrieved by
            // getCurrentKey getCurrentValue methods
            return true;
        }
    }
	  
	  
	 */

	@Override
	public boolean nextKeyValue() throws IOException, InterruptedException {
		
		if (key == null) {
            key = new LongWritable();
        }
        key.set(pos);
        
        
        String famID=this.readFamilyID();
        
        //verbose = famID.equals("iORTHO026253");
        
        while (famID.isEmpty()){
        	return false; //terminate sequence
        }
        value = new GeneFamily(famID);
        
        
        try {
		    value.setNewick(this.readNewickString());
		    /*if (verbose){
		    	System.out.println(value.getNewick());
		    }*/
		    int numGenes=this.readNumberOfGenes();
		    /*if (verbose){
		    	System.out.println(numGenes);
		    }*/
		    for (int i=0; i<numGenes; i++){
		    	value.addGeneSeq(this.readGene(),this.readSequence());
		    	
		    	
		    }
		    
        } catch (IOException e){
        	System.out.println(e.getMessage());
        	return false;
        }
        
        
        
        
        return true;
	}
	
	
	private String readFamilyID() throws IOException {
		return processLine();
	}
	
	private String readNewickString() throws IOException{
		String line=processLine();
		if (line.isEmpty())
			throw new IOException("Newick String was empty!");
		return line;

	}
	
	private int readNumberOfGenes() throws IOException {
		String line=processLine();
		if (line.isEmpty())
			throw new IOException("Number of species missing!");
		return Integer.parseInt(line);
	}
	
	private Gene readGene() throws IOException {
		String line=processLine();
		String [] strs = line.split("\t");
		if (strs.length!=2){
			System.out.println(line);
			for (String s : strs){
				System.out.println(s);
			}
			throw new IOException("Gene Org pair not read correctly!");
		}
		return new Gene(strs[0],strs[1]);
	}
	

	private BaseSequence readSequence() throws IOException{
		String line=processLine();
		if (line.isEmpty())
			throw new IOException("BaseSequence is empty!");
		return new BaseSequence(line);
	}
	
	private String processLine() throws IOException {
    	
		/*if (verbose){
			System.out.println("processLine");
		}*/
		
    	Text totalText = new Text();
    	Text tempText = new Text();
    	int newSize=0;
    	
    	/*if (verbose){
        	System.out.println(pos+"\t"+end);
        	System.out.println(totalText);
        }*/
        while (pos < end) { //sometimes one line is multiple maxLineLengths until \n is reached
            newSize = in.readLine(tempText, maxLineLength,Math.max(maxBytesToConsume(end-pos),maxLineLength));
            totalText.append(tempText.getBytes(),0, tempText.getLength());
            
            
            tempText.clear();

            pos += newSize; //update position in split
            if (newSize < maxLineLength) { //is this line the last one before '\n'?
                break;
            }
            
            /*if (verbose){
            	System.out.println(pos+"\t"+end);
            	System.out.println(totalText);
            }*/
        }

        
        return totalText.toString();
    }   
}

